'''
csv_to_custom_json by n4n5
'''
from csv import reader as CSVreader
from os.path import isfile
from json import dumps as JSONstringify

# pylint: disable=W0102
# -> dangerous arguments
# pylint: disable=C0103
# -> Function name "parseFile"


def parseFile(path_to_file="", schema=None, options_user={}):
    """ Global function to parse a file """
    def check_options(options_user, attr, default_value):
        """ Check options or put default value """
        if attr in options_user:
            return options_user[attr]
        return default_value
    options = {
        "arrayParse": check_options(options_user, "arrayParse", True),
        "callBackForce": check_options(options_user, "callBackForce", False),
        "debug": check_options(options_user, "debug", False),
        "error": check_options(options_user, "error", False),
        "lineCallBack": check_options(options_user, "lineCallBack", None),
        "parse": check_options(options_user, "parse", True),
        "separator": check_options(options_user, "separator", ","),
        "overrideFirstLine": check_options(options_user, "overrideFirstLine", False),
        "avoidVoidLine": check_options(options_user, "avoidVoidLine", False)
    }
    if options["debug"]:
        if isinstance(schema, (list, dict)) and schema is not None:
            print("HAS SCHEMA")
        else:
            print("NO SCHEMA")
            print("OPTIONS", JSONstringify(options))
        if options["error"] == "no":
            print("Useless informations : just use try catch if you don't want error :)")
    if isinstance(path_to_file, str) and not isfile(path_to_file):
        if options["error"] == "no":
            return []
        raise ValueError(f"Can't access to the file : '{path_to_file}'")

    if isinstance(path_to_file, str):
        csv_file = open(path_to_file, encoding="utf-8")
        line_reader = CSVreader(csv_file, delimiter=options["separator"])
    elif isinstance(path_to_file, list):
        line_reader = path_to_file

    def create_fields_binding(schema_object, first_line, start_path=[]):
        """ Create fields bindings """
        bindings = []
        for index, value in enumerate(schema_object):
            is_list = isinstance(schema_object, list)
            is_dict = isinstance(schema_object, dict)
            if is_list:
                one_element = index
            elif is_dict:
                one_element = value
            path = []
            if len(start_path) != 0:
                path = start_path.copy()
            path.append(f"{one_element}")
            if isinstance(schema_object[one_element], (dict, list)):
                if isinstance(schema_object[one_element], list):
                    bindings.append({
                        "name": one_element,
                        "path": path,
                        "type": "helper-array"
                    })
                bindings = bindings + \
                    create_fields_binding(
                        schema_object[one_element], first_line, path)
            else:
                if is_list and options["arrayParse"] and schema_object[one_element] in first_line:
                    bindings.append({
                        "name": schema_object[one_element],
                        "path": path,
                        "value": "string"
                    })
                else:
                    if one_element in first_line or callable(schema_object[one_element]):
                        bindings.append({
                            "name": one_element,
                            "path": path,
                            "value": schema_object[one_element]
                        })
                    else:
                        bindings.append({
                            "name": one_element,
                            "path": path,
                            "type": "static",
                            "value": schema_object[one_element]
                        })
        return bindings

    def parse_line(line, rows, first_line):
        """" Parse one line """
        if isinstance(schema, list):
            obj = []
        else:
            obj = {}
        all_values = line
        for one_row in rows:
            one_path_row = one_row["path"]
            one_path_name = one_row["name"]
            all_path = one_path_row
            current_value = None
            if ('type' not in one_row) or ('type' in one_row and one_row["type"] is None):
                if 'value' not in one_row:
                    schema_value = None
                else:
                    schema_value = one_row["value"]
                if one_row["name"] in first_line:
                    index = first_line.index(one_row["name"])
                else:
                    index = -1
                if index == -1:
                    current_value = schema_value
                else:
                    if index < len(all_values):
                        current_value = all_values[index]
                if options["parse"] and current_value is not None:
                    if schema_value == "int" and current_value != '':
                        current_value = int(current_value)
                    elif schema_value == "float":
                        current_value = float(current_value)
                    elif schema_value == "string":
                        current_value = str(current_value)
                    elif callable(schema_value):
                        if callable(current_value):
                            # When the value is in an array
                            current_value = schema_value(all_values)
                        else:
                            current_value = schema_value(current_value)
            elif ('type' in one_row and one_row["type"] == "helper-array"):
                current_value = []
            elif ('type' in one_row and one_row["type"] == "static"):
                current_value = one_row["value"]
            good_place = None
            if len(all_path) > 1:
                good_place = obj
                long = len(all_path)
                for count in range(0, long):
                    next_path = all_path[count]
                    if isinstance(good_place, list):
                        next_path_int = int(next_path)
                    if count == (long - 1):
                        if isinstance(good_place, dict):
                            good_place[next_path] = ""
                    else:
                        if (isinstance(good_place, list) and next_path_int not in good_place) or next_path not in good_place:
                            if isinstance(good_place, list):
                                if len(good_place) < (next_path_int+1):
                                    # len() returns 0 and the first index of the list is 0 !
                                    good_place.insert(next_path_int, {})
                            else:
                                good_place[next_path] = {}
                        if isinstance(good_place, list):
                            if next_path_int < len(good_place):
                                good_place = good_place[next_path_int]
                        else:
                            good_place = good_place[next_path]
                if isinstance(good_place, list):
                    good_place.append(current_value)
                elif isinstance(good_place, dict):
                    good_place[one_path_name] = current_value
            else:
                if isinstance(obj, list):
                    place = int(one_path_row[0])
                    obj.insert(place, current_value)
                elif isinstance(obj, dict):
                    obj[one_path_row[0]] = current_value
        return obj

    def parse_first_line(first_line):
        """ Parse the first line """
        if schema is not None:
            # None is default value for schema
            cols = create_fields_binding(schema, first_line)
            if options["debug"]:
                print("BINDINGS:", JSONstringify(
                    cols, default=lambda o: '<not serializable>'))
        else:
            def dupli(element):
                """" Duplicate the first line """
                return {
                    "name": element,
                    "path": [element]
                }
            cols = [dupli(x) for x in first_line]
        return cols

    def reader(line_reader):
        """" Read the file """
        final_json = []
        if isinstance(path_to_file, str):
            first_line = next(line_reader)
        elif isinstance(path_to_file, list):
            first_line = line_reader[0].split(options["separator"])
            line_reader = line_reader[1:]
        if isinstance(options["overrideFirstLine"], list):
            first_line = options["overrideFirstLine"]
        rows = parse_first_line(first_line)
        for one_line in line_reader:
            parsed_line = {}
            if isinstance(path_to_file, list):
                one_line = one_line.split(options["separator"])
            elif isinstance(path_to_file, str) and isinstance(one_line, list) and not one_line:
                one_line = ['']  # create a fake void line
            if options["avoidVoidLine"]:
                if (isinstance(one_line, list) and not one_line) or (isinstance(one_line, list) and len(one_line) >= 1 and one_line[0] == "") or one_line == "" or one_line == "\n" or one_line == "\r\n":
                    continue
            parsed_line = parse_line(one_line, rows, first_line)
            if callable(options["lineCallBack"]):
                res_callback = options["lineCallBack"](parsed_line, one_line)
                if res_callback is None:
                    if options["callBackForce"]:
                        parsed_line = res_callback
                    else:
                        if options["debug"]:
                            print(
                                "CallBack force at False and callBack result is not correct")
                else:
                    parsed_line = res_callback
            final_json.append(parsed_line)
        return final_json
    converted = reader(line_reader)
    if isinstance(path_to_file, str):
        csv_file.close()
    return converted



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_debug(self):
        test = parseFile("./tests/not_found.csv", None, {
            "debug": True,
            "error": "no",
        })
        self.assertEqual(test, [])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_array_parse(self):
        def function1(allValues):
            mystring = 'toto{}'.format(','.join(map(str, allValues)))
            return mystring

        def function2(uselessArg):
            return "arrow"
        test = parseFile("./tests/simple_complexe.csv", [
            "num4",
            "text",
            function1,
            function2
        ], {
            "arrayParse": False
        })
        self.assertEqual(test, [
            [
                "num4",
                "text",
                "toto1,2,3,4",
                "arrow"
            ],
            [
                "num4",
                "text",
                "toto4,5,6,7",
                "arrow"
            ],
            [
                "num4",
                "text",
                "toto7,8,9,10",
                "arrow"
            ]
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_override_first_line(self):
        newFirstLine = ["hello1", "hello2", "hello3", "hello4"]

        def func(value):
            return 'The number 4 is {}'.format(value)
        test = parseFile("./tests/simple_complexe.csv", {
            "num4": {
                "num4": "string",
                "hello1": "string",
                "hello2": "int",
                "hello3": "float",
                "hello4": func
            }
        }, {
            "overrideFirstLine": newFirstLine
        })
        self.assertEqual(test, [
            {
                "num4": {
                    "num4": "string",
                    "hello1": "1",
                    "hello2": 2,
                    "hello3": 3,
                    "hello4": "The number 4 is 4"
                }
            },
            {
                "num4": {
                    "num4": "string",
                    "hello1": "4",
                    "hello2": 5,
                    "hello3": 6,
                    "hello4": "The number 4 is 7"
                }
            },
            {
                "num4": {
                    "num4": "string",
                    "hello1": "7",
                    "hello2": 8,
                    "hello3": 9,
                    "hello4": "The number 4 is 10"
                }
            }
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_complex_json_10(self):
        def function1(allValues):
            mystring = 'toto{}'.format(','.join(map(str, allValues)))
            return mystring

        def function2(uselessArg):
            return "arrow"
        test = parseFile("./tests/simple_complexe.csv", [
            "num4",
            "text",
            function1,
            function2,
            {
                "staticValue": "value",
                "num1": "int"
            }
        ])
        self.assertEqual(test, [
            [
                "4",
                "text",
                "toto1,2,3,4",
                "arrow",
                {
                    "staticValue": "value",
                    "num1": 1
                }
            ],
            [
                "7",
                "text",
                "toto4,5,6,7",
                "arrow",
                {
                    "staticValue": "value",
                    "num1": 4
                }
            ],
            [
                "10",
                "text",
                "toto7,8,9,10",
                "arrow",
                {
                    "staticValue": "value",
                    "num1": 7
                }
            ]
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_callBack_force_2(self):
        def function1(values):
            return None

        def function2(useless, useless2):
            return None
        test = parseFile("./tests/simple.csv", {
            "num1": "string",
            "num2": function1,
            "num3": "string"
        }, {
            "callBackForce": False,
            "lineCallBack": function2,
            "debug": True,
        })
        self.assertEqual(test, [
            {'num1': '1', 'num2': None, 'num3': '3'},
            {'num1': '4', 'num2': None, 'num3': '6'},
            {'num1': '7', 'num2': None, 'num3': '9'},
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_array_parse(self):
        def function1(allValues):
            mystring = 'toto{}'.format(','.join(map(str, allValues)))
            return mystring

        def function2(uselessArg):
            return "arrow"
        test = parseFile([
            "num1,num2,num3,num4",
            "1,2,3,4",
            "4,5,6,7",
            "",
            "7,8,9,10"
        ], [
            "num4",
            "text",
            [
                "arrayHereLol",
                [
                    "andHereLol",
                    {
                        "obj": "lol",
                        "num4": "int"
                    }
                ]
            ],
            function1,
            function2,
            {
                "staticValue": "value",
                "num1": "int"
            }
        ], {
            "avoidVoidLine": True
        })
        self.assertEqual(test, [
            [
                "4",
                "text",
                [
                    "arrayHereLol",
                    [
                        "andHereLol",
                        {
                            "obj": "lol",
                            "num4": 4
                        }
                    ]
                ],
                "toto1,2,3,4",
                "arrow",
                {
                    "staticValue": "value",
                    "num1": 1
                }
            ],
            [
                "7",
                "text",
                [
                    "arrayHereLol",
                    [
                        "andHereLol",
                        {
                            "obj": "lol",
                            "num4": 7
                        }
                    ]
                ],
                "toto4,5,6,7",
                "arrow",
                {
                    "staticValue": "value",
                    "num1": 4
                }
            ],
            [
                "10",
                "text",
                [
                    "arrayHereLol",
                    [
                        "andHereLol",
                        {
                            "obj": "lol",
                            "num4": 10
                        }
                    ]
                ],
                "toto7,8,9,10",
                "arrow",
                {
                    "staticValue": "value",
                    "num1": 7
                }
            ]
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_callBack_force_2(self):

        def function2(parsedLine, useless2):
            name = list(parsedLine.keys())
            tab = {}
            tab[name[0]] = parsedLine[name[0]] + "1"
            tab[name[1]] = parsedLine[name[1]] + "1"
            tab[name[2]] = parsedLine[name[2]] + "1"
            return tab
        test = parseFile("./tests/simple.csv", {
            "num1": "string",
            "num2": "string",
            "num3": "string"
        }, {
            "callBackForce": True,
            "lineCallBack": function2
        })
        self.assertEqual(test, [
            {
                "num1": "11",
                "num2": "21",
                "num3": "31"
            },
            {
                "num1": "41",
                "num2": "51",
                "num3": "61"
            },
            {
                "num1": "71",
                "num2": "81",
                "num3": "91"
            }
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_debug(self):
        file = "./tests/not_found.csv"
        try:
            test = parseFile(file, None, {
                "debug": True,
            })
        except Exception as e:
            self.assertEqual(
                e.__str__(), f"Can't access to the file : '{file}'")



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_complex_json_8(self):
        def function1(allValues):
            mystring = 'toto{}'.format(','.join(map(str, allValues)))
            return mystring

        def function2(uselessArg):
            return "arrow"
        test = parseFile("./tests/simple_complexe.csv", {
            "num1": [
                "num4",
                "text",
                function1,
                function2
            ]
        })
        self.assertEqual(test, [
            {
                "num1": [
                    "4",
                    "text",
                    "toto1,2,3,4",
                    "arrow"
                ]
            },
            {
                "num1": [
                    "7",
                    "text",
                    "toto4,5,6,7",
                    "arrow"
                ]
            },
            {
                "num1": [
                    "10",
                    "text",
                    "toto7,8,9,10",
                    "arrow"
                ]
            }
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_complex_json_3(self):
        test = parseFile("./tests/simple_complexe.csv", {
            "hello": {
                "uno": {
                    "dos": {
                        "tres": {
                            "num4": "string"
                        }
                    }
                }
            },
            "bonjour": {
                "un": {
                    "deux": {
                        "trois": {
                            "num2": "string",
                            "num1": "int"
                        }
                    }
                }
            },
            "num3": ""
        })
        self.assertEqual(test, [
            {
                "hello": {
                    "uno": {
                        "dos": {
                            "tres": {
                                "num4": "4"
                            }
                        }
                    }
                },
                "bonjour": {
                    "un": {
                        "deux": {
                            "trois": {
                                "num2": "2",
                                "num1": 1
                            }
                        }
                    }
                },
                "num3": "3"
            },
            {
                "hello": {
                    "uno": {
                        "dos": {
                            "tres": {
                                "num4": "7"
                            }
                        }
                    }
                },
                "bonjour": {
                    "un": {
                        "deux": {
                            "trois": {
                                "num2": "5",
                                "num1": 4
                            }
                        }
                    }
                },
                "num3": "6"
            },
            {
                "hello": {
                    "uno": {
                        "dos": {
                            "tres": {
                                "num4": "10"
                            }
                        }
                    }
                },
                "bonjour": {
                    "un": {
                        "deux": {
                            "trois": {
                                "num2": "8",
                                "num1": 7
                            }
                        }
                    }
                },
                "num3": "9"
            }
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_custom_separator(self):
        test = parseFile("./tests/simple_customSeparator.csv", {
            "num1": "",
            "num2": "",
            "num3": ""
        }, {
            "separator": ";"
        })
        self.assertEqual(test, [
            {
                "num1": "1",
                "num2": "2",
                "num3": "3"
            },
            {
                "num1": "4",
                "num2": "5",
                "num3": "6"
            },
            {
                "num1": "7",
                "num2": "8",
                "num3": "9"
            }
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_complex_json_2(self):
        test = parseFile("./tests/simple_complexe.csv", {
            "num1": {
                "num4": {
                    "num4": "string"
                }
            },
            "num2": "",
            "num3": ""
        })
        self.assertEqual(test, [
            {
                "num1": {
                    "num4": {
                        "num4": "4"
                    }
                },
                "num2": "2",
                "num3": "3"
            },
            {
                "num1": {
                    "num4": {
                        "num4": "7"
                    }
                },
                "num2": "5",
                "num3": "6"
            },
            {
                "num1": {
                    "num4": {
                        "num4": "10"
                    }
                },
                "num2": "8",
                "num3": "9"
            }
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_parse_value(self):
        test = parseFile("./tests/simple.csv", {
            "num1": "int",
            "num2": "float",
            "num3": "string"
        })
        self.assertEqual(test, [
            {
                "num1": 1,
                "num2": 2,
                "num3": "3"
            },
            {
                "num1": 4,
                "num2": 5,
                "num3": "6"
            },
            {
                "num1": 7,
                "num2": 8,
                "num3": "9"
            }
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_callBack_force(self):

        def function2(useless, useless2):
            return  # return nothing
        test = parseFile("./tests/simple.csv", {
            "num1": "string",
            "num2": "string",
            "num3": "string"
        }, {
            "lineCallBack": function2
        })
        self.assertEqual(test, [
            {
                "num1": "1",
                "num2": "2",
                "num3": "3"
            },
            {
                "num1": "4",
                "num2": "5",
                "num3": "6"
            },
            {
                "num1": "7",
                "num2": "8",
                "num3": "9"
            }
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_complex_json_7(self):
        def function1(allValues):
            mystring = 'toto{}'.format(','.join(map(str, allValues)))
            return mystring

        def function2(uselessArg):
            return "arrow"

        def function3(allValues):
            mystring = 'hey{}'.format(','.join(map(str, allValues)))
            return mystring
        test = parseFile("./tests/simple_complexe.csv", {
            "hello": [
                {
                    "num4": "int",
                    "num1": "string"
                }
            ],
            "hello2": [
                {
                    "num4": "int",
                    "num3": function3,
                    "num1": [
                        "num4",
                        "text",
                        function1,
                        function2
                    ]
                }
            ]
        })
        self.assertEqual(test, [
            {
                "hello": [
                    {
                        "num4": 4,
                        "num1": "1"
                    }
                ],
                "hello2": [
                    {
                        "num4": 4,
                        "num3": "hey3",
                        "num1": [
                            "4",
                            "text",
                            "toto1,2,3,4",
                            "arrow"
                        ]
                    }
                ]
            },
            {
                "hello": [
                    {
                        "num4": 7,
                        "num1": "4"
                    }
                ],
                "hello2": [
                    {
                        "num4": 7,
                        "num3": "hey6",
                        "num1": [
                            "7",
                            "text",
                            "toto4,5,6,7",
                            "arrow"
                        ]
                    }
                ]
            },
            {
                "hello": [
                    {
                        "num4": 10,
                        "num1": "7"
                    }
                ],
                "hello2": [
                    {
                        "num4": 10,
                        "num3": "hey9",
                        "num1": [
                            "10",
                            "text",
                            "toto7,8,9,10",
                            "arrow"
                        ]
                    }
                ]
            }
        ])



import unittest

from src.csv_to_custom_json import parseFile


class Test(unittest.TestCase):
    def test_complex_json_6(self):
        test = parseFile("./tests/simple_complexe.csv", {
            "hello": [
                {
                    "num4": "int",
                    "num1": "string"
                }
            ],
            "hello2": [
                [
                    {
                        "num3": "string"
                    }
                ]
            ]
        })
        self.assertEqual(test, [
            {
                "hello": [
                    {
                        "num4": 4,
                        "num1": "1"
                    }
                ],
                "hello2": [
                    [
                        {
                            "num3": "3"
                        }
                    ]
                ]
            },
            {
                "hello": [
                    {
                        "num4": 7,
                        "num1": "4"
                    }
                ],
                "hello2": [
                    [
                        {
                            "num3": "6"
                        }
                    ]
                ]
            },
            {
                "hello": [
                    {
                        "num4": 10,
                        "num1": "7"
                    }
                ],
                "hello2": [
                    [
                        {
                            "num3": "9"
                        }
                    ]
                ]
            }
        ])
